(*----------------------------------------------------------------------(C)-*)
(* Copyright (C) 2006 -2012 Konstantin Korovin and The University of Manchester.
   This file is part of iProver - a theorem prover for first - order logic.

   iProver is free software: you can redistribute it and / or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   iProver is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with iProver. If not, see < http:// www.gnu.org / licenses />. *)
(*----------------------------------------------------------------------[C]-*)

open Lib
open Git_info
open Options
open Statistics
open Printf
open Bmc1InitTarget

open Logic_interface
open Problem_properties

open Proof_search_schedule

(* record backtrace for debugging          *)
(* compile with -g option to get the trace *)



(*----- debug modifiable part-----*)

let dbg_flag = false

type dbg_gr = 
  | D_elapsed
  | D_preprocess
  | D_trace
  | D_trace_prep
  | D_symb_reachability 
  | D_parse
  | D_sub_typing
  | D_axs_distinct 
  | D_axs_less_range
  | D_axs_eq
  | D_sem_filter
  | D_clausify

let dbg_gr_to_str = function 
  | D_elapsed -> "elapsed"
  | D_preprocess -> "preprocess"	
  | D_trace -> "trace"
  | D_trace_prep -> "trace_preprocess"
  | D_symb_reachability -> "symb_reach"
  | D_parse -> "parse"
  | D_sub_typing -> "sub_typing"
  | D_axs_distinct  ->  "axs_distinct"
  | D_axs_less_range -> "axs_less_range"
  | D_axs_eq -> "axs_eq"
  | D_sem_filter -> "sem_filter"
  | D_clausify -> "clausify"

let dbg_groups =
  [
   (* D_trace_prep; *)
   D_elapsed;
(* D_trace; *)
(*   D_symb_reachability; *)
(* D_parse; *)
(*   D_sub_typing; *)
(*   D_clausify;*)
(*   D_axs_eq *)
 ]
    
let module_name = "iprover"

(*----- debug fixed part --------*)

let () = Lib.dbg_flag_msg dbg_flag module_name

let dbg group str_lazy =
  Lib.dbg_out_pref dbg_flag dbg_groups group dbg_gr_to_str module_name str_lazy

let dbg_env group f =
  Lib.dbg_env_set dbg_flag dbg_groups group f

(*----- debug -----*)


exception Time_out_real = Signals.Time_out_real
exception Time_out_virtual = Signals.Time_out_virtual
 
(*------------Out source info -----------------------------*)
let source_info_str = 
  let git_pref = "git:" in
  "\n"^(s_pref_str())^" iProver source info "^"\n\n"^
(*  git_pref^(" date: "^(git_info.git_date))^"\n"^ *)
  git_pref^(" sha1: "^(git_info.git_sha1))^"\n"^
  git_pref^" non_committed_changes: "^(string_of_bool git_info.git_non_committed_changes)^"\n"^
  git_pref^" last_make_outside_of_git: "^(string_of_bool !make_outside_git)^"\n" 
												
  
let out_source_info () = 
  out_str source_info_str

(*-----------------------------------------*)

let result_handler_prep result =

  (if (val_of_override !current_options.bmc1_incremental)
  then
    (* note that in the case of bmc1 here only preprocess results are handled; *)
    (* results during bmc1 search are handled in bmc1_loop  *)
    Bmc1_loop.result_handler_bmc1_preprocess result
  else
    result_handler_basic result
  );
  (if 
    ((not (val_of_override !current_options.bmc1_incremental)) 
   || (val_of_override !current_options.bmc1_out_stat <> BMC1_Out_Stat_None))
  then 
    out_stat ()
  )
    

let result_handler result =

  result_handler_basic result;

  (if 
    ((not (val_of_override !current_options.bmc1_incremental)) 
   || (val_of_override !current_options.bmc1_out_stat <> BMC1_Out_Stat_None))
  then 
    out_stat ()
  )
    
(*-------------------------------------------------*)
(* elapsed time code *)
(*-------------------------------------------------*)

(* keep the time *)
let last_timestamp = ref 0.0

(* set the timestamp *)
let timestamp () = last_timestamp := Unix.gettimeofday ()

(* helper: print the elapsed time, keep the time stamp *)
let elapsed_helper status =
  (* current time *)
  let current = Unix.gettimeofday () in
  (* report *)
  out_str (Format.sprintf "Timer report: %s: elapsed time %.3fs" status (current -. !last_timestamp));
  (* return current time *)
  current

(* print the elapsed time, keep the time stamp *)
let elapsed status = ignore (elapsed_helper status)

(* print the elapsed time, reset timer *)
let elapsed_ts status = last_timestamp := (elapsed_helper status)

(*----------------Top Function--------------------*)

(*-- run_iprover: initialises, ----------------*)
(*-- parses and then runs main on the prepocessed clauses------*)

let run_iprover () =

  (*-------------------*)
  out_source_info ();
  out_str(options_to_str !current_options);
  
  let filtered_out_clauses_ref = ref [] in
  
  (try

    (*---------Set System Signals--------------------*)

    Signals.set_sys_signals ();
    Signals.set_time_out ();
    
    (*---------------Parse the input-------------*)

      
    (* we need to switch off type checking during parsing since vars are retyped during parsing *)
    (* restore after parsing *)

    let input_symbol_type_check = !current_options.symbol_type_check in

    !current_options.symbol_type_check <- false;
    
    (*------------- parsing ------------------*)

    dbg D_trace_prep (lazy ("parsing start"));
    dbg_env D_elapsed (fun () -> timestamp ());
    assign_float_stat (get_time_fun 3 ParseFiles.parse) parsing_time;
    dbg_env D_elapsed (fun () -> (elapsed_ts "parsing and AIG"));
    !current_options.symbol_type_check <- input_symbol_type_check;
    
(*    (if !current_options.clausify_out
    then
      (
*)
     dbg_env D_clausify
	 (
	  fun () ->
	    (
	     let clause_list = Parser_types.get_clauses_without_extra_axioms () in
	     let (epr, non_epr) = List.partition Clause.is_epr clause_list in
	     out_str ("% "^pref_str^"Clauses after clausification:\n\n");
	     out_str ("% "^pref_str^"EPR clauses:\n\n");
	     Clause.out_clause_list_tptp epr;
	     out_str ("\n\n"^"% "^pref_str^"non-EPR clauses:\n\n");
	     Clause.out_clause_list_tptp non_epr;
	     out_str "\n\n";
(*	     exit(0);*)
	    )
	 );
(*
    else ()
    );
*)
    (* tsar: do MC preprocessing: $traget, $init *)
    (* may be move to the end; after normal preprocessing *)

(* bmc1_preprocess_input -> bmc1_input_transformation *)

    dbg D_trace_prep (lazy ("bmc1_input_transformation start"));
    Bmc1_loop.bmc1_input_transformation (); 
    dbg_env D_elapsed (fun () -> (elapsed_ts "bmc1_input_transformation"));

    (*---- after parsing we need to calculate has_conj_symb/has_non_prolific_conj_symb ----------*)
  
    dbg D_trace_prep (lazy ("change_conj_symb_input start"));
    Parser_types.change_conj_symb_input ();
    dbg_env D_elapsed (fun () -> (elapsed_ts "change_conj_symb_input"));

    (*------ calculate symbol reachability based on father of map --------*)

    if !current_options.bmc1_symbol_reachability then      
      (
       let reach_map_start_time = Unix.gettimeofday () in
        dbg D_trace_prep (lazy ("symbol_reach start"));
       (* symbol depth is assigned during Symbol_reach.symbol_reach *)
       let symb_reach_map = (Symbol_reach.symbol_reach !(Parser_types.neg_conjectures)) in
	   
       dbg_env D_symb_reachability
	 (fun () ->
	   Symbol_reach.out_symb_reach_map
	     (symb_reach_map);
	   
	   out_str (Format.sprintf
		      "%sTime for symbol reachability: %.3fs@\n@."
		      pref_str
		      (Unix.gettimeofday () -. reach_map_start_time)
		   );
	 );       
    dbg_env D_elapsed (fun () -> (elapsed_ts "symbol_reach"));
      );
    
    
    (*-------------------------------*)
    let current_clauses = ref (Parser_types.get_clauses_without_extra_axioms ()) in
    (*-------------------------------*)

    (*---debug-------*)
    dbg_env D_parse 
	(fun () -> 
	  Logic_interface.out_symbs ();
	  Logic_interface.out_terms ();
	  Clause.out_clauses_param (Clause.CL_List !current_clauses)
	);
   
    dbg D_trace_prep (lazy ("get_bv_axioms start"));
    let bv_axioms = Eq_axioms.get_bv_axioms () in
    current_clauses := List.rev_append bv_axioms !current_clauses;
    dbg_env D_elapsed (fun () -> (elapsed_ts "get_bv_axioms"));


    (* At the moment Parsed_input_to_db.input_clauses_ref are not memory released! *)
    (* we can replace Parsed_input_to_db.input_clauses_ref with *)
    (* global  Parsed_input_to_db.current_clauses, which are gradually replaced by preprocessing but should be carefull how intput caluses are used below: finite_models eq_axioms etc. *)

    (* tsar: AIG does not contains equalities, so unflatten is inapplicable *)
    (if (not !current_options.aig_mode) && !current_options.prep_unflatten
    then (
      dbg D_trace_prep (lazy ("unflatten start"));
      current_clauses := Inference_rules.unflatten !current_clauses;
      dbg_env D_elapsed (fun () -> (elapsed_ts "unflatten"));
    ));

    
(* TODO: make preprocessing that does not depend on prop_solver first *)

(* we can not init prop_solver before subtyping bacause terms for grounding depend on types! 
 

(*-------------------------------*)
    Prop_solver_exchange.init_solver_exchange ();
 (*-------------------------------*)
    List.iter Prop_solver_exchange.add_clause_to_solver !current_clauses;

(* temporarily unassign_is_essential_input_symb before preprocessing *)	
    unassign_is_essential_input_symb !current_clauses; 
      (* preprocess befroe eq axioms, otherwise can be expensive *)


      current_clauses := Preprocess.preprocess ~is_eq_problem:(has_equality !current_clauses) !current_clauses;  

      assign_is_essential_input_symb !current_clauses;

      dbg D_preprocess
	(lazy (" Clauses after preprocessing: "^(Clause.clause_list_to_tptp !current_clauses)^"\n"));

*)


    (*-------sybtyping-------------*)		

(* need to assign is essential input before subtyping *)
(*
    Clause.assign_is_essential_input_symb (Clause.CL_List !current_clauses);
    
    let sub_type_is_on () =
      !current_options.sub_typing
	&&
      (not (val_of_override !current_options.bmc1_incremental))
	&&
      (match !current_options.schedule with
      | Schedule_verification_epr
      | Schedule_verification_epr_tables
      | Schedule_verification_aig
	-> false
      | _ -> true
      )
    in
    
    (if (sub_type_is_on ())
    then
      (       
       (* we need to unflatten x!= y before subtyping *)
       (* if it was on in the options unflattening was applied above *)
       (if (not !current_options.prep_unflatten)
       then
	 current_clauses := Inference_rules.unflatten !current_clauses
       else ());  
       current_clauses := Type_inf.sub_type_inf !current_clauses;)
    else ());
    
    dbg D_sub_typing (lazy "-----------After Subtyping:---------");	
    dbg D_sub_typing (lazy (Clause.clause_list_to_tptp !current_clauses));	
    dbg D_sub_typing (lazy "------------------------------------");


*)
    dbg D_trace_prep (lazy ("init_solver_exchange start"));
    (*-------------------------------*)
    Prop_solver_exchange.init_solver_exchange ();
    (*-------------------------------*)
    dbg_env D_elapsed (fun () -> (elapsed_ts "init_solver_exchange"));

    dbg D_trace_prep (lazy ("add_clause_to_solver start"));
    List.iter Prop_solver_exchange.add_clause_to_solver !current_clauses;
    dbg_env D_elapsed (fun () -> (elapsed_ts "add_clause_to_solver"));
    
    (* with sat_mode one should be careful with options!*)
    (* switch off resolution! *)
    (* out_str ("\n\n!!!!! Switch from Finite Models to SAT mode!!!\n\n");
       if !current_options.sat_mode
       then
       (current_options:= (named_option_finite_models ()).options;
       sat_mode !current_clauses )
       else
     *)
 
(*----------------------- Preprocessing ------------------------*)
   

      (* temporarily unassign_is_essential_input_symb before preprocessing *)	
      dbg D_trace_prep (lazy ("unassign_is_essential_input_symb start"));
      Clause.unassign_is_essential_input_symb (Clause.CL_List !current_clauses); 
      dbg_env D_elapsed (fun () -> (elapsed_ts "unassign_is_essential_input_symb"));
      (* preprocess befroe eq axioms, otherwise can be expensive *)

      dbg D_trace_prep (lazy ("preprocess start"));
      current_clauses := Preprocess.preprocess ~is_eq_problem:(has_equality !current_clauses) !current_clauses;  
      dbg D_trace_prep (lazy ("preprocess finish"));
      dbg_env D_elapsed (fun () -> (elapsed_ts "preprocess"));

      Clause.assign_is_essential_input_symb (Clause.CL_List !current_clauses); 

      dbg D_preprocess
	(lazy (" Clauses after preprocessing: "^(Clause.clause_list_to_tptp !current_clauses)^"\n"));


    (*-------sybtyping-------------*)		

(* need to assign is essential input before subtyping *)
    
    let sub_type_is_on () =
      !current_options.sub_typing
	&&
      (not (val_of_override !current_options.bmc1_incremental))
	&&
      (match !current_options.schedule with
      | Schedule_verification_epr_old
      | Schedule_verification_epr_tables
      | Schedule_verification_epr
	-> false
      | _ -> true
      )
    in
    
    (if (sub_type_is_on ())
    then
      (       
       (* we need to unflatten x!= y before subtyping *)
       (* if it was on in the options unflattening was applied above *)
       (if (not !current_options.prep_unflatten)
       then
	 current_clauses := Inference_rules.unflatten !current_clauses
       else ());  
       current_clauses := Type_inf.sub_type_inf !current_clauses;)
    else ());
    
    dbg D_sub_typing (lazy "-----------After Subtyping:---------");	
    dbg D_sub_typing (lazy (Clause.clause_list_to_tptp !current_clauses));	
    dbg D_sub_typing (lazy "------------------------------------");

    Prop_solver_exchange.init_gr_by (); 

    Clause.assign_is_essential_input_symb (Clause.CL_List !current_clauses);

(*---------------- Extra axioms -----------------*)

(*--- axs distinct ---*)
      let distinct_ax_list = Eq_axioms.distinct_ax_list () in

      dbg_env D_axs_distinct  
	( fun () ->
	  
      	  out_str "\n-----------Distinct Axioms:---------\n";
	  out_str ((Clause.clause_list_to_tptp distinct_ax_list)^"\n\n");
	  out_str "\n--------------------\n";
	 );

      (* Clauses are input clauses *)
      Clause.assign_is_essential_input_symb (Clause.CL_List distinct_ax_list);
      
      current_clauses := List.rev_append distinct_ax_list !current_clauses;

(*--- axs less/range ---*)
      
      let less_range_axioms = Eq_axioms.less_range_axioms () in
     
      dbg_env D_axs_less_range 
	( fun () ->
	out_str "\n-----------Less Range Axioms:---------\n";
	out_str ((Clause.clause_list_to_tptp less_range_axioms)^"\n\n");
	out_str "\n--------------------\n";
	 );
	  
      current_clauses := List.rev_append less_range_axioms !current_clauses;

      (* Clauses are input clauses *)      
      Clause.assign_is_essential_input_symb (Clause.CL_List less_range_axioms);

      

(*-------- Equality Axioms ----------------------*)
      
      let gen_equality_axioms = ref [] in

      
      (if !current_options.pure_diseq_elim
      then 
	(current_clauses:=  Eq_axioms.pure_dis_eq_elim !current_clauses)
      );

(*  current_clauses_no_eq used in sat *)
      let current_clauses_no_eq = ref (!current_clauses) in
	
      (if !current_options.brand_transform
      then
	(
	 dbg_env D_axs_eq
	   (fun () ->
	     out_str "\n-----------Before Brand transform:---------\n";
	     out_str ((Clause.clause_list_to_tptp !current_clauses)^"\n\n");
	     out_str "\n--------------------\n";
	   );
	 
	 current_clauses := (Eq_axioms.eq_axioms_flatting !current_clauses);
	 
	 dbg_env D_axs_eq
	   (fun () ->
	     out_str "\n-----------After Brand transform:---------\n";
	     out_str ((Clause.clause_list_to_tptp !current_clauses)^"\n\n");
	     out_str "\n--------------------\n";
	   )	  
	)
      else
	(	 
	 gen_equality_axioms := Eq_axioms.eq_axiom_list !current_clauses;
	 
	 dbg_env D_axs_eq
	      (fun () ->
		
		out_str "\n-----------Eq Axioms:---------\n";
		out_str ((Clause.clause_list_to_tptp !gen_equality_axioms)^"\n\n");
		out_str "\n--------------------\n";
		
	      );	    
	 current_clauses := List.rev_append !gen_equality_axioms !current_clauses
	)
      );
    
      
      (*--------------semantic filter---------------------------*)
      
      (* out_str "\n\n\n Before get_side_clauses \n\n\n "; *)
      (* side_clauses used in sem filter                   *)

      let get_side_clauses () =
	if val_of_override !current_options.bmc1_incremental then
	  (* we initialised Bmc1Axioms.init_bound !current_clauses above *)
	  (* simulate is true in increment_bound *)
    (* tsar: TODO: replace with proper init axioms *)
	  [](*((Bmc1Axioms.get_bound_assumptions 1)
	   @(Bmc1Axioms.increment_bound 0 1 false true))
    *)
	else []
      in
      
      (*     out_str "\n\n\n Before Filtering \n\n\n ";*)
      
      if !current_options.prep_sem_filter_out
      then
	(
	 
	 (*-------------------------------------------------*)
	 out_str (pref_str^"Semantic Filtering...\n");
	 (*-------------------------------------------------*)
	 
	 (*	 out_str (pref_str^"Semantically Preprocessed Clauses:\n");*)
	 
	 let side_clauses = (get_side_clauses ()) in
	 let filtered_clauses =
	   Prep_sem_filter_unif.sem_filter_unif (!current_clauses) side_clauses
	 in
	 let filtered_in_clauses =
	   filtered_clauses.Prep_sem_filter_unif.filtered_in in
	 
	 out_str (pref_str^"Input cluases sem filter:\n");
	 Clause.out_clause_list_tptp !current_clauses;
	 
	 out_str ("\n\n"^pref_str^"Semantically Preprocessed Clauses:\n");
	 
	 Clause.out_clause_list_tptp filtered_in_clauses;
	 
	 (*
	   out_str ("\n\n"^pref_str^"Filtered in clauses: \n"^(Clause.clause_list_to_string filtered_in_clauses));
	   out_str ("\n\n"^pref_str^"Filtered out clauses: \n"
	           ^(Clause.clause_list_to_string filtered_clauses.Prep_sem_filter_unif.filtered_out));
	  *)
	 
	 out_str "\n\n";
	 out_str (unknown_str ());
	 out_stat ();
	 exit(0);
	 (* raise SZS_Unknown *)
	)
      else
	(
	 (* if (!current_options.prep_sem_filter &&
	    (not (Symbol.is_input Symbol.symb_equality)))
	  *)
	 (* was as above but equality should be ok, *)
	 
	 (* problem with bmc1 *)
	 
	 if ((!current_options.prep_sem_filter != Sem_Filter_None)
	       &&
	     (not (val_of_override !current_options.bmc1_incremental))
	       &&
	     (match !current_options.schedule with
	     | Schedule_verification_epr_old
	     | Schedule_verification_epr_tables
	     | Schedule_verification_epr
	       -> false
	     | _ -> true
	     )
	       
	    )
	 then
	   (
	    (*-------------------------------------------------*)
	    out_str (pref_str^"Semantic Filtering...\n");
	    (*-------------------------------------------------*)
	    
	    (*	  out_str "\n\n\n!!!! Fix Sem Filter for Finite models and BMC1 !!!!!!\n\n\n";*)
	    (*          current_clauses := Prep_sem_filter.filter !current_clauses)*)
	    (*	  current_clauses := List.sort cmp_clause_length !current_clauses;*)
	    
	    let side_clauses = get_side_clauses () in
	    let filtered_clauses =
	      Prep_sem_filter_unif.sem_filter_unif !current_clauses side_clauses
	    in
	    current_clauses := filtered_clauses.Prep_sem_filter_unif.filtered_in;
	  
	    filtered_out_clauses_ref :=
	      filtered_clauses.Prep_sem_filter_unif.filtered_out;
	    (
	     if (!current_options.sat_mode || !current_options.sat_finite_models ||
	     !current_options.schedule = Schedule_sat)
	     then
	       (

		let filtered_clauses =
		  Prep_sem_filter_unif.sem_filter_unif
		    !current_clauses_no_eq (List.rev_append !gen_equality_axioms side_clauses)
		in
		
		current_clauses_no_eq :=
		  filtered_clauses.Prep_sem_filter_unif.filtered_in;

		filtered_out_clauses_ref :=
		  filtered_clauses.Prep_sem_filter_unif.filtered_out;

	       )
	     else ()
	    )
	   )
	 else ()
	);
      
      (*--------------End sem filter---------------------------*)

  
      (*debug *)
      (*
	out_str "\n\n Clauses before proving\n\n";
	out_str (Clause.clause_list_to_tptp !current_clauses);
       *)
      
      (
       if (!current_options.dbg_out_stat) (* && (not !input_problem_props.epr) *)
       then 

	 (* get statistics on non-cyclic/epr subtyping without taking into account pure EPR *)

	 ignore (Finite_models.init_fm_state  !current_clauses_no_eq);
      );

      (if (!current_options.dbg_out_stat)
      then
	failwith "Output satistics";
      );

  
	 
      (*-------------------------------------------------*)
      out_str (pref_str^"Proving...");
      (*-------------------------------------------------*)

      dbg D_trace (lazy ("In run_iprover: input size = "^(string_of_int(List.length !current_clauses))));
   
      (* Clause.out_clause_list !current_clauses; *)
      

      let finite_model_clauses_ref = current_clauses_no_eq in



(*----- problem properties --------*)

      let problem_properties = (Problem_properties.get_problem_props !current_clauses) in 
      Problem_properties.problem_props_to_statistics problem_properties;      
      out_str (pref_str^"Problem Properties \n");
      out_str ("\n"^(Problem_properties.problem_props_to_string problem_properties)^"\n");


(*----- solver call before schedules ------*)

(* solver is called first time in the  Proof_search_schedule.schedule_run  *)

(*------------------------ Proving part ---------------------*)

      let schedule_clauses =
      {
        Proof_search_schedule.proof_clauses = !current_clauses;
        Proof_search_schedule.finite_model_clauses = !finite_model_clauses_ref;
        Proof_search_schedule.filtered_out_clauses = !filtered_out_clauses_ref;
      }
      in

      let sched_run sched = 
        if (val_of_override !current_options.bmc1_incremental)
        then
          (* create initial MC state *)
          let state = Bmc1_loop.bmc1_init_solver_state sched schedule_clauses in
          (* run BMC1 loop on that state *)
          if !current_options.bmc1_ucm
          then
            Bmc1_loop.bmc1_mp_loop_init state
          else
            Bmc1_loop.bmc1_loop state
        else
          Proof_search_schedule.schedule_run schedule_clauses sched
    in
      
      (if (not !current_options.instantiation_flag) && (not !current_options.resolution_flag)
      then
	failwith "No solver is selected: see --instantiation_flag, --resolution_flag"
      );

      let result =
	begin	  
          match !current_options.schedule with
          | Schedule_default ->
              sched_run (Proof_search_schedule.default_schedule (problem_properties))

	  | Schedule_sat ->
              sched_run (Proof_search_schedule.sat_schedule (problem_properties))
		
	  | Schedule_none ->
             (* let prover_functions = Proof_search_loop.create_provers_current_options !current_clauses in*)
	      sched_run (Proof_search_schedule.trivial_schedule ())
		
(*            Proof_search_loop.full_loop prover_functions ~time_limit:Undef current_clauses *)
	      	
          | Schedule_verification_epr_old ->
              sched_run (Proof_search_schedule.verification_epr_schedule_old ())
		
          | Schedule_verification_epr_tables ->
              sched_run (Proof_search_schedule.verification_epr_schedule_tables ())	                  

          | Schedule_verification_epr ->
              sched_run (Proof_search_schedule.verification_epr_schedule ())
	end
      in
      dbg D_trace (lazy "Normal: exit with result");
      result_handler result 

  with
    
(* preprocessing results *)
  

  | Unsatisfiable_gr 
  | Unsatisfiable_gr_na
    ->
      dbg D_trace (lazy "Prep: Unsat");
      result_handler_prep (Res_prop_solver_unsat)      
     
  | Empty_clause (clause) ->
      
      dbg D_trace (lazy "Prep: Empty");
      result_handler_prep (Res_empty_clause (clause))
      
	
  | Discount.Satisfiable all_clauses
    ->
      (
       dbg D_trace (lazy "Prep: sat discount");
       result_handler_prep (Res_resolution_sat (all_clauses, !filtered_out_clauses_ref))
      )
	
  | Instantiation.Satisfiable all_clauses
    ->
      (
       dbg D_trace (lazy "Prep: sat instantiation");
       result_handler_prep (Res_instantiation_sat (all_clauses, !filtered_out_clauses_ref))
      )

  | Termination_Signal ->
      (kill_all_child_processes ();
       out_str (unknown_str ());
       out_str "\n Termination Signal\n";
       
       (* Do not output statistics in BMC1 mode with
	  -- bmc1_out_stat none *)
       if (not (val_of_override !current_options.bmc1_incremental)) ||
       (not (val_of_override !current_options.bmc1_out_stat =
	     BMC1_Out_Stat_None)) then
	 out_stat ())
	
  | Time_out_real ->
      (kill_all_child_processes ();
       out_str (unknown_str ());
       out_str "Time Out Real\n";
       
       (* Do not output statistics in BMC1 mode with
	  -- bmc1_out_stat none *)
       if (not (val_of_override !current_options.bmc1_incremental)) ||
       (not (val_of_override !current_options.bmc1_out_stat =
	     BMC1_Out_Stat_None)) then
	 out_stat()
	   
      )
	
  | Time_out_virtual ->
      (kill_all_child_processes ();
       out_str (unknown_str ());
       out_str "Time Out Virtual\n";
       
       (* Do not output statistics in BMC1 mode with
	  -- bmc1_out_stat none *)
       if (not (val_of_override !current_options.bmc1_incremental)) ||
       (not (val_of_override !current_options.bmc1_out_stat =
	     BMC1_Out_Stat_None)) then
	 out_stat())
	
  | Proof_search_schedule.Schedule_Terminated ->
      (kill_all_child_processes ();
       out_str (unknown_str ());
       out_str "Schedule_Terminated:  try an extended schedule or with an unbounded time limit";
       
       (* Do not output statistics in BMC1 mode with
	  -- bmc1_out_stat none *)
       if (not (val_of_override !current_options.bmc1_incremental)) ||
       (not (val_of_override !current_options.bmc1_out_stat =
	     BMC1_Out_Stat_None)) then
	 out_stat ())
	
	(* Silently terminate after BMC1 maximal bound proved *)
  | Exit ->
       dbg D_trace (lazy "Exit: done with BMC");
	
  | x ->
      
      (       
       if !current_options.dbg_backtrace then
	 Format.eprintf "Unexpected error after main.Backtrace:@\n%s@\n@." (Printexc.get_backtrace ());
       
       kill_all_child_processes ();
       out_str (unknown_str ());
       out_stat ();
       raise x)
  )


let _ = run_iprover ()



(*---------------------------Commented----------------------------*)
