(*----------------------------------------------------------------------(C)-*)
(* Copyright (C) 2006-2012 Konstantin Korovin and The University of Manchester. 
    This file is part of iProver - a theorem prover for first-order logic.

	iProver is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
      the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
  iProver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
    See the GNU General Public License for more details.
	You should have received a copy of the GNU General Public License
	    along with iProver.  If not, see <http://www.gnu.org/licenses/>.         *)
(*----------------------------------------------------------------------[C]-*)

 

open Options
open Statistics 
open Lib

open Logic_interface
open PredElim


let top_term = Parser_types.top_term


(*----- debug modifiable part-----*)

let dbg_flag = false

type dbg_gr = 
  | D_init
  | D_out_prep_clauses
  | D_solver 
  | D_marshal
  | D_cone_symb
  | D_cone_symb_full

let dbg_gr_to_str = function 
  | D_init -> "init"
  | D_out_prep_clauses -> "out_prep_clauses"
  | D_solver -> "solver"
  | D_marshal -> "marshal"
  | D_cone_symb -> "cone_symb"
  | D_cone_symb_full -> "cone_symb_full"
	
let dbg_groups =
  [
(*   D_init; *)
(*   D_out_prep_clauses;*)
(*   D_marshal; *)
   D_solver;
   D_cone_symb;
(*   D_cone_symb_full; *)
 ]
    
let module_name = "preprocess"

(*----- debug fixed part --------*)

let () = dbg_flag_msg dbg_flag module_name

let dbg group str_lazy = 
  Lib.dbg_out_pref dbg_flag dbg_groups group dbg_gr_to_str module_name str_lazy

let dbg_env group f = 
  Lib.dbg_env_set dbg_flag dbg_groups group f
    
(*----- debug -----*)


let is_ver_epr () = 
  (!current_options.aig_mode || (val_of_override  !current_options.bmc1_incremental)) 
    

let remove_true_false clause = 
  let changed = ref false in
  let f lit_rest lit = 
    let pred = Term.lit_get_top_symb lit in 
    let sign = Term.is_pos_lit lit in 
    let is_true_atom = (pred == Symbol.symb_true) in
    let is_false_atom = (pred == Symbol.symb_false) in
    match (sign, is_true_atom, is_false_atom) with
(* true lit *)
    | (true, true, _) | (false,_,true) -> 
	raise Eliminated
	  
(* false lit *)
    | (false,true,_) | (true,_,true) ->
	changed:= true;
	lit_rest

    |_-> 
	lit::lit_rest
  in
  let new_lits = Clause.fold f [] clause in
  if !changed 
  then 
    let tstp_source = Clause.tstp_source_true_false clause in
    let new_clause = create_clause tstp_source new_lits in 
    check_empty_clause new_clause; 
    new_clause
  else
    clause

(* can raise Eliminated *)
let get_sim_self_fun_list () = 
  [

(* self simplifications *)
(* Simplify.equality_resolution;*) (* since preprocess also applited to eq axioms this doe not work *)

   remove_true_false;
   Simplify.tautology_elim;
   
(*--prop subs-*)
   if !current_options.prep_gs_sim  then Simplify.forward_prop_subsume else id_fun;
 ]
    
    
let prop_simp clause_list = 
  (* debug *)
  (* List.iter 
     (fun c -> 
     Format.printf "Prep: %a@.\n" (TstpProof.pp_clause_with_source false) c; 	
     Prop_solver_exchange.add_clause_to_solver c;
     )
     clause_list;
   *)
  List.iter 
    Prop_solver_exchange.add_clause_to_solver clause_list;
  (if ((Prop_solver_exchange.solve ()) = PropSolver.Unsat)
  then 
    ((* Format.eprintf "Unsatisfiable after solve call in Preprocess.prop_sim@."; *)
     (* Raise separate exception, since BMC1 must continue if
	simplified and must not continue if solver is in invalid state *)
     (* raise PropSolver.Unsatisfiable *)
     raise Unsatisfiable_gr
    )
  else ());
  let simplify_clause rest clause = 
    (Prop_solver_exchange.prop_subsumption clause)::rest
  in
  List.fold_left simplify_clause [] clause_list
    

   

(*------Non-equational to Equational (based on input options)-----------*) 

module SymbKey = 
  struct
    type t    = symbol
    let equal = (==)
    let hash  = Symbol.get_fast_key 
  end 

module PredToFun = Hashtbl.Make (SymbKey)
    

let pred_to_fun_symb pred_to_fun_htbl pred = 
  try 
    PredToFun.find pred_to_fun_htbl pred
  with 
    Not_found ->
      let new_symb_name = ("$$iProver_FunPred_"^(Symbol.get_name pred)) in
      let new_type = 
	match (Symbol.get_stype_args_val pred) with
	|Def(old_args, old_val) ->
	    Symbol.create_stype old_args Symbol.symb_bool_type
	|Undef -> 
	    Symbol.create_stype [] Symbol.symb_default_type    
      in
      let fun_symb = 
	Symbol.create_from_str_type_property
	  new_symb_name new_type Symbol.FunPred in
      let added_fun_symb = SymbolDB.add_ref fun_symb symbol_db_ref in
      PredToFun.add pred_to_fun_htbl pred added_fun_symb;
      added_fun_symb

	
let pred_to_fun_atom pred_to_fun_htbl atom =
  match atom with 
  |Term.Fun (pred,args,_) -> 
      if not (pred == Symbol.symb_typed_equality)
      then 
	let fun_symb = pred_to_fun_symb pred_to_fun_htbl pred in
	let fun_term = add_fun_term_args fun_symb args in 
	let eq_term  = add_typed_equality_sym Symbol.symb_bool_type fun_term top_term in
	eq_term
      else
	atom
  |_ -> failwith "pred_to_fun_atom should not be var"


let pred_to_fun_lit pred_to_fun_htbl lit =
  let new_lit = Term.apply_to_atom (pred_to_fun_atom pred_to_fun_htbl) lit in 
  TermDB.add_ref new_lit term_db_ref
    
    

let pred_to_fun_clause pred_to_fun_htbl clause = 
  let new_lits = 
    List.map
      (pred_to_fun_lit pred_to_fun_htbl)
      (Clause.get_literals clause) in
  let tstp_source = Clause.tstp_source_non_eq_to_eq clause in		
  let new_clause = create_clause tstp_source new_lits in
  (* Clause.assign_non_eq_to_eq_history new_clause clause; *)
  
  new_clause



(* *)
let res_prep_options () = 
  {!current_options
  with 
   (*----Resolution---------*)
   
   resolution_flag = true;
   
   res_prop_simpl_new = 
   if (is_ver_epr ()) 
   then 
     false
   else
     true;
   res_prop_simpl_given = 
   if (is_ver_epr ()) 
   then 
     false
   else
     true;
   res_passive_queue_type = PQT_PriorityQueues;
   res_passive_queues =
   [
    [Cl_Num_of_Lits false; Cl_Num_of_Symb false]
  ];
   res_passive_queues_freq = [150];
   
   res_forward_subs = Subs_Full;
   res_backward_subs = Subs_Full;
   res_forward_subs_resolution = true;
   (*  res_forward_subs_resolution    = true; exp later for sat *)
   (* res_backward_subs_resolution   = false; *)
   res_backward_subs_resolution = true;
   res_time_limit = 60.0;
 }
    
    
let res_prep clause_list = 
  let old_options = !current_options in 
  current_options := res_prep_options ();
  let module ResInput =
    struct
      let res_module_name = "Res prep"
      let input_clauses = clause_list
      let is_res_prepocessing = true 
    end 
  in
  let module ResM = Discount.Make (ResInput) in
  let new_clauses = ResM.res_prep () in 
  ResM.clear_all ();
  current_options := old_options;
  new_clauses
    
    
let preprocess ~is_eq_problem clause_list =
  print_string ((s_pref_str ())^"Preprocessing...");
  flush stdout;
  (* Tsar: to implement preprocessing for AIG/verification *)

  (if (is_ver_epr ())
  then
    (
     let verific_options = {!current_options with

			    prep_gs_sim = false;
			    res_sim_input = false;
			    prep_res_sim = false;

			    (* symbol_type_check = !current_options.symbol_type_check; *)
			    symbol_type_check = false;
			    prep_sem_filter = Sem_Filter_None;
			    (* uncomment the following line to improve output for the DEMO *)
			    (* inst_out_proof = ValueCmd false; *)
			  } in
     current_options := verific_options;
    )
  );

  let current_list = ref clause_list in
  (if !current_options.non_eq_to_eq 
  then 
    (
     let pred_to_fun_htbl = PredToFun.create (SymbolDB.size !symbol_db_ref) in
     current_list := 
       (List.map (pred_to_fun_clause pred_to_fun_htbl) !current_list)
    )
  else ()
  ); 

  let sim_self_fun_list = get_sim_self_fun_list () in
  let simpl_sim rest c = 
    try
      let new_c = fold_left_fun_list c sim_self_fun_list in   
      new_c::rest
    with 
      Eliminated -> 
	rest
  in 
  current_list:= List.fold_left simpl_sim [] !current_list;
  
(*
  (if  !current_options.prep_gs_sim 
  then 
  current_list := prop_simp !current_list
  else ());
 *)
(*  current_list := List.map unflatten !current_list;  
    out_str (" \n after unflat: \n"^(Clause.clause_list_to_string !current_list));
 *)
  
  (if !current_options.prep_res_sim 
  then
    current_list := res_prep !current_list
  else ()
  );
  (if !current_options.pred_elim 
  then
    (

     let (init_pred_elim_set, _not_used_num_occur_map) = 
       PredElim.get_most_preds_to_eliminate (Clause.CL_List (!current_list)) in     
     
     

     let cmp_num_occ num_occur_map p1 p2 = 
       try
	 let get_occ_num pred = 
	   (SMap.find pred num_occur_map)
	 in
	 Pervasives.compare (get_occ_num p1)  (get_occ_num p2)
       with 
	 Not_found -> 0 
	     (* if one of them of the pred is not found
		(it can happen when elimination set contains predicates that are not in the map/simplified cluase set) *) 
     in       

(* order by aig depth *)
(* high depth first which corresponds to inputs (invert depth in preprocess in reasoning or not ?) *)

     let cmp_depth_inv_fun p1 p2 = Pervasives.compare (get_pred_depth p2) (get_pred_depth p1) in

     let pred_elim_cmp_fun context = 
       let (pred_elim_set, num_occur_map) = 
	 PredElim.get_most_preds_to_eliminate (Clause.CL_Context (context)) in     
       

       (if (val_of_override !current_options.bmc1_incremental)
       then 
	 lex_combination [cmp_depth_inv_fun;  (cmp_num_occ num_occur_map)] 
       else
	 (cmp_num_occ num_occur_map)
       )

     in

     let num_of_input_clauses = (Statistics.get_val_stat num_of_input_clauses) in

     let pred_elim_options = 
       {
	pe_has_eq = is_eq_problem; 
	pe_estim_num_of_cl =  
	(if (is_ver_epr ()) 
	then
	  if  (num_of_input_clauses < 1000000 (* 100 000 *))
	  then
	    45 
	  else
	    25  (*260*) (* 25 *) (* 45 *) (* 65 *) (*85*) (* 65 *)
	else 
	  40 (* 1000 *)
	);
	
	pe_clause_length_limit = 
	(if  (is_ver_epr ())  
	then
	  13  (* 8 *) (* 26 *) (* 13*)	      
	else 
	  8 (* 8 *) (* 50 *)
	);

(* not evaluated on aigs *)
	pe_keep_elim = 
	(if (is_ver_epr ())  
	then
	  (fun ~num_cl_before ~num_cl_after -> 
	    (num_cl_after <= num_cl_before)  || (num_cl_after <= 2)) (* (num_cl_after <= 6) ) *)
	else
	  (fun ~num_cl_before ~num_cl_after -> 
	    (num_cl_after <= num_cl_before) (* ||  (num_cl_after <= 6)*) )
	);
	pe_elim_order_cmp_fun = pred_elim_cmp_fun;
	pe_elimination_set = init_pred_elim_set;
	
(* simplifications *)

(*	subs_cl_to_cl_limit = 100000; *)
	subs_cl_to_cl_limit = 200000;  


(* sim prop *)
	prop_glb_subs =  
	(if (is_ver_epr ())   
	then	    
	  if (num_of_input_clauses < 10000 )
	  then
	    true  (* prop global subsumtion changed for finite models exp. *)
	  else
	    false
	else
	  (* false*)  true 
	);
	
(* sim local *) 
	
	(* Options.res_subs_type: type res_subs_type = Subs_Full | Subs_Subset | Subs_By_Length of int *)
	
	lcl_fwd_subs = 
	(if  (is_ver_epr ()) 
	then 
	  if (num_of_input_clauses < 400000)
	  then
	    true
	  else
	    false   
	else 
	  true
	);

	lcl_fwd_subs_res = 
	(if  (is_ver_epr ()) 
	then	      
	  if (num_of_input_clauses < 100000 (* 100000*))
	  then
	    true
	  else
	    false
	else
	true (* false *)
	);
	
	lcl_bwd_subs     = 
	(if (is_ver_epr ())  
	then
	  if  (num_of_input_clauses < 1000000) (* (num_of_input_clauses < 400000)*)
	  then
	    Subs_By_Length(1) (* Subs_By_Length(1) *)
	  else
	    Subs_Subset
	      
	      (* Subs_Subset*)  (*(Subs_By_Length(1)) *) (* Subs_Full *)
	else 
	 Subs_Subset (* (Subs_By_Length(2)) *) (* (Subs_Full) *)
	);

	lcl_bwd_subs_res =  (* Subs_Full;*)  
	(if  (is_ver_epr ()) 
	then	    
	  if (num_of_input_clauses < 100000 (* 100000*))
	  then
	    Subs_By_Length(1)
	  else
	    Subs_Subset
	      
	      (* Subs_Subset*) (*  (Subs_By_Length(1)) *)(* Subs_Full *)
	else 
	  Subs_Subset (* (Subs_By_Length(2))*) (* (Subs_Full) *)
	);

	
(* sim global *)
	
	glb_fwd_subs =
	(if (is_ver_epr ()) 
	then
	  if (num_of_input_clauses < 200000 (*200000 *))
	  then
	    true 
	  else
	    false
	else
	  true  (* false *)
	);
	
	glb_fwd_subs_res =
	(if (is_ver_epr ()) 
	then
	  if (num_of_input_clauses < 100000 (* 100000 *) )
	  then
	    true
	  else
	    false
	else
	  false (* false;*)
	);

	glb_bwd_subs     = 
	(if (is_ver_epr ()) 
	then
	  if (num_of_input_clauses < 1000000 (*200000*))
	  then
	    Subs_By_Length(1)
	  else
	    Subs_Subset (*  Subs_By_Length(1) *)  (*Subs_Full*)
	else 
	  Subs_Subset (* Subs_By_Length(1);*) (* Subs_By_Length(4); *) (*Subs_Full; *)	  
	);
	
	glb_bwd_subs_res =
	(if (is_ver_epr ()) 
	then
	  if (num_of_input_clauses < 1000000 (* 200000 *))
	  then
	    Subs_By_Length(1)
	  else
	    Subs_Subset  (* Subs_By_Length(1) *)(* Subs_Full*)
	else 
	  Subs_Subset (* Subs_By_Length(1);*) (* Subs_By_Length(4); *) (*Subs_Full; *)	  
	);
	
      }
     in
     (*let fixed_point_reached = ref false in*)
     (* out_str (Clause.clause_list_to_string !current_list); *)

     let new_clauses = PredElim.predicate_elimination pred_elim_options (Clause.CL_List (!current_list)) in
     current_list:= new_clauses;
(*
  while (not !fixed_point_reached)
  do
(* optimisation for runs after first: make a flag whether pred_clauses got modified;*)
(* if  at the moment of elimination of a pred its pred_clauses was not modified we can immediatly abort this elimination *)
(* also the same pe_state can be used just need to reset the elimination queue *)

  let new_clauses = PredElim.predicate_elimination pred_elim_options (Cl_List (!current_list)) in
  Statistics.incr_int_stat 1 Statistics.pred_elim_cycles;
  let reduction = (List.length !current_list) - (List.length new_clauses)  in
  out_dbg ~g:1 (lazy ("------------------"));
  out_dbg ~g:1 (lazy ("has_eq: "^(string_of_bool pred_elim_options.pe_has_eq)));
  out_dbg ~g:1 (lazy ("before predelim: "^(string_of_int (List.length !current_list))));
  out_dbg ~g:1 (lazy (Clause.clause_list_to_string !current_list));
  out_dbg ~g:2 (lazy ("after predelim: "^(string_of_int (List.length new_clauses))
  ^" reduction: "^(string_of_int reduction)));
  out_dbg ~g:2 (lazy  (Clause.clause_list_to_string new_clauses));      
  out_dbg ~g:2 (lazy ("------------------"));
  fixed_point_reached:= reduction = 0;
  current_list:= new_clauses;
  done     
 *)
    )
  );	
  (match !current_options.ground_splitting with
  |Split_Input |Split_Full ->    
      let split_result = 
	(Splitting.ground_split_clause_list !current_list) in
      incr_int_stat 
	(Splitting.get_num_of_splits split_result) num_of_splits; 
      incr_int_stat 
	(Splitting.get_num_of_split_atoms split_result) num_of_split_atoms;	
      current_list:=Splitting.get_split_list split_result
  |Split_Off-> ()
  );  	
  out_str "\n";
  dbg_env D_out_prep_clauses
    (
     fun () -> (
       let clause_list = !current_list in
       let (epr, non_epr) = List.partition Clause.is_epr clause_list in
       out_str ("% "^pref_str^"Clauses after preprocessing: "^(string_of_int (List.length clause_list))^"\n\n");
       out_str ("% "^pref_str^"EPR clauses: "^(string_of_int (List.length epr))^"\n\n");
       Clause.out_clause_list_tptp epr;
       out_str ("\n\n"^"% "^pref_str^"non-EPR clauses:"^(string_of_int (List.length non_epr))^" \n\n");
       Clause.out_clause_list_tptp non_epr;
       out_str "\n\n";
      )
    );

(*--------- resets solver --------*)
  (
   if (!current_options.reset_solvers)
   then
     (
      dbg D_solver (lazy "start: reset");
      
      Prop_solver_exchange.reset_solvers (); 
      
      dbg D_solver (lazy "end: reset");

      dbg D_solver (lazy "start: add clauses");
      let start_time = Unix.gettimeofday () in
      
      List.iter 
	Prop_solver_exchange.add_clause_to_solver !current_list;
      
      let end_time = Unix.gettimeofday () in
      dbg D_solver (lazy ("end: add clauses: "^(string_of_float (end_time -. start_time))^"\n"));
     )
   else
     ()
  );
  
  dbg_env D_marshal
    (fun () ->
      dbg D_marshal (lazy ("start: Marshal clauses\n"));
      let start_time = Unix.gettimeofday () in
      let bytes = Marshal.to_bytes !current_list [] in
      let end_time = Unix.gettimeofday () in
      dbg D_marshal (lazy ("end: Byte length"^(string_of_int (Bytes.length bytes))
			   ^" time: "^((string_of_float (end_time -. start_time)))
			   ^"\n"))
    );

  dbg_env D_cone_symb
    (fun () ->
      
      let is_relevant_symb symb = 
	(not (symb == Symbol.symb_ver_next_state)) 	  
      in    
      dbg D_cone_symb (lazy ("start full_rel: init clauses: "^(string_of_int (List.length !current_list))^"\n"));
      let start_time = Unix.gettimeofday () in
      let full_rel = Cone_symb.create_full_rel_cl_list ~is_relevant_symb !current_list in 
      let end_time = Unix.gettimeofday () in
      dbg D_cone_symb (lazy ("end full_rel: time "^((string_of_float (end_time -. start_time)))));
      
      let depth_0_symb_set = SSet.of_list [Symbol.symb_ver_property] in
      dbg D_cone_symb (lazy ("start compute cone: "));
      let start_time = Unix.gettimeofday () in
      let cone = Cone_symb.compute_cone full_rel ~terminating_symb_set:SSet.empty ~depth_0_symb_set in 
      let end_time = Unix.gettimeofday () in
      let depth =  (* 4*)  -1  in (* all reach clauses *) 
      let cone_clauses = CSet.elements (Cone_symb.get_cone_clauses cone ~depth) in 
      dbg D_cone_symb (lazy ("end compute cone: size "^(string_of_int (List.length cone_clauses))
			^" time "^((string_of_float (end_time -. start_time)))));
      Cone_symb.out_cone ~symbs:false  ~clauses:false ~stats:true cone;

      out_warning ("preprocess cone-reduced clause set restricted to depth "^(string_of_int depth)^"\n\n ");
      current_list:= cone_clauses; 
 
    );


(*--------- cone bc_imp_inh ---------*)

  (
   if (bc_imp_inh_exists !current_options BCI_conj_cone) 
   then
     (* TODO move non-prolific *)
     
     let symb_fill_num_of_occ_clause occ_smap cl = 
       let f (curr_occ_smap, counted_sset) symb =  (* each symb is counted only once per cluase*)
	 if (not (SSet.mem symb counted_sset))
	 then 
	   try 
	     let old_num_occ = SMap.find symb curr_occ_smap in 
	     let new_smap = SMap.add symb (old_num_occ+1) curr_occ_smap in 
	     let new_counted_sset = SSet.add symb counted_sset in 
	     (new_smap, new_counted_sset)
	   with 
	     Not_found ->
	       (SMap.add symb 1 curr_occ_smap, SSet.add symb counted_sset)
	 else
	   (curr_occ_smap, counted_sset) 
       in
       let (fill_occ_smap,_counted) =  Clause.fold_sym f (occ_smap,SSet.empty) cl in 
       fill_occ_smap
     in
     let symb_num_of_occ_map = List.fold_left symb_fill_num_of_occ_clause SMap.empty !current_list in

     (* get into list and order by priority *)
     dbg_env D_cone_symb 
       (fun () ->
	 let symb_num_occ_list =
	   SMap.fold (fun symb n rest -> ((symb, n):: rest)) symb_num_of_occ_map []
	 in
	 let sorted_num_occ_list =
	   List.sort (fun (_, d1) (_, d2) -> compare d1 d2) symb_num_occ_list 
	 in	
	 out_str ("num occ symbols: "^(string_of_int (List.length sorted_num_occ_list))^"\n\n");
	 dbg_env D_cone_symb_full 
	   (fun () ->
	     List.iter
	       (fun (symb, depth) ->
		 out_str ((Symbol.to_string symb)^": "^(string_of_int depth)))
	       sorted_num_occ_list
	   )
       );

(*
     let prolific_frac = 0.1 in 

     let num_of_clauses = List.length !current_list in 
     let prolific_bound =  (int_of_float (prolific_frac *. (float_of_int num_of_clauses))) in
     let is_prolific symb = 
       try 
	 let num_of_occ = SMap.find symb symb_num_of_occ_map in	 
	 dbg_env D_cone_symb 
	   (fun () -> 
	     if  prolific_bound <= num_of_occ	
	     then
	       dbg D_cone_symb (lazy ("prolific symb: "^(Symbol.to_string symb)^" num occ: "^(string_of_int num_of_occ)^"\n"))
	   );

	 prolific_bound <= num_of_occ	  
       with
	 Not_found -> false
     in

 
   
     let is_relevant_symb symb = 
       (((not (is_special_symbol symb)) && (not (is_prolific symb)))
       || (symb == Symbol.symb_ver_property)) in    
*)

     let pred_symb_only = false in (* false: all symbols not just predicates for general problems; for bmc1 ok true *)

     let tolerance = !current_options.conj_cone_tolerance in

     let is_relevant_symb symb = 
       ((not (is_special_symbol symb))
      || (symb == Symbol.symb_ver_property)) 
     in  
     
     dbg D_cone_symb (lazy
			("start full_rel: init clauses: "
			 ^(string_of_int (List.length !current_list))^"\n"));

     let start_time = Unix.gettimeofday () in
     let full_rel = Cone_symb.create_full_rel_cl_list ~tolerance ~symb_num_of_occ_map ~is_relevant_symb ~pred_symb_only  !current_list in 
     let end_time = Unix.gettimeofday () in
     dbg D_cone_symb (lazy ("end full_rel: time "^((string_of_float (end_time -. start_time)))));    

(* get conjecture relevant symbols *)
     let get_clause_rel_symbs clause = 
       if pred_symb_only
       then
	 Clause.find_all_pred ~is_relevant_pred:(fun _sign symb -> is_relevant_symb symb) clause 
       else
	 Clause.find_all_sym ~is_relevant_symb clause 
     in 

     let get_conj_cl_rel_symbs sset clause = 
       SSet.union (get_clause_rel_symbs clause) sset
     in
     let conj_preds = List.fold_left get_conj_cl_rel_symbs SSet.empty !(Parser_types.neg_conjectures) in
     
(*    let depth_0_symb_set = SSet.of_list [Symbol.symb_ver_property] in *)
     let depth_0_symb_set = SSet.add Symbol.symb_ver_property conj_preds in 
     dbg D_cone_symb (lazy ("start compute cone: "));
     let start_time = Unix.gettimeofday () in
     let cone = Cone_symb.compute_cone full_rel ~terminating_symb_set:SSet.empty ~depth_0_symb_set in 
     let end_time = Unix.gettimeofday () in
     dbg D_cone_symb (lazy ("end compute cone: time "^((string_of_float (end_time -. start_time)))));

     dbg_env D_cone_symb
       (fun () -> Cone_symb.out_cone ~symbs:false ~clauses:false ~stats:true cone;);
 
     dbg_env D_cone_symb_full 
       (fun () -> Cone_symb.out_cone ~symbs:true  ~clauses:true ~stats:true cone;);
     
(* asssign bc_imp_inh to clauses based on depth *)
     let bc_imp_inh_shift = get_bc_imp_inh_shift !current_options BCI_conj_cone in
(*    dbg D_cone (lazy (("bc_imp_inh_shift: ")^(string_of_int bc_imp_inh_shift )^"\n")); *)
     let symb_depth_map = Cone_symb.get_cone_symb_depth_map cone in
     let get_symb_depth symb =        
       try
	 SMap.find symb symb_depth_map
       with 
	 Not_found -> max_int 
     in
     let symb_to_clauses_full_rel = Cone_symb.get_symb_to_clauses full_rel in
     let is_trigger symb clause =
       try
	 let cset = SMap.find symb symb_to_clauses_full_rel in 
	 CSet.mem clause cset
       with
	 Not_found -> false
     in
       
     let get_min_symb_depth clause = 
       let f curr_min symb =
	 if (is_trigger symb clause)
	 then
	   let symb_depth = get_symb_depth symb in
	   if curr_min < symb_depth 
	   then 
	     curr_min
	   else
	     symb_depth
	 else
	   curr_min
       in
       Clause.fold_sym f max_int clause
     in
     let assign_bc_imp_inh clause = (* min depth of its preds *)
       let cl_depth = get_min_symb_depth clause in 
       if cl_depth = max_int 
       then ()
       else 
	 (
	  let c_importance = (cl_depth + bc_imp_inh_shift) in 
	  dbg D_cone_symb_full (lazy ((Clause.to_string clause)^" cl_depth: "^(string_of_int cl_depth)^"\n"));
	  Clause.assign_bc_imp_inh clause c_importance;
	 )
     in
     let depth = -1 in
     let cone_clauses = (Cone_symb.get_cone_clauses cone ~depth) in
     CSet.iter assign_bc_imp_inh cone_clauses 
(* note we do not assigne cone clauses to current_list but just changing the priority of them *)
   else  
     (dbg D_cone_symb (lazy (" BCI_conj_cone is not in the !current_options \n"));)
  );

  !current_list
    
