open Lib 
open Statistics
open Options
open Logic_interface

(*---------------Probelm Properties---------------------*)

type problem_properties =
    {
     mutable epr : bool;
     mutable horn : bool;
     mutable has_eq : bool;
   }

type problem_props = problem_properties

(* initialisation is like this due to how epr/horn is reassigned *)
let empty_problem_props () =
  {
   epr = true;
   horn = true;
   has_eq = false;
 }


let problem_props_to_string props =
  let props_list =
    [
     ("EPR", string_of_bool props.epr);
     ("Horn", string_of_bool props.horn);
     ("Has equality", string_of_bool props.has_eq)
   ]
  in
  Options.opt_val_list_to_str props_list

let problem_props_to_statistics props = 
  assign_int_stat (bool_to_int props.epr) is_epr;
  assign_int_stat (bool_to_int props.horn) is_horn;
  assign_int_stat (bool_to_int props.has_eq) has_eq

let get_problem_props clause_list =
  let props = empty_problem_props () in
  let f cl =
    (if props.epr then
      props.epr <- Clause.is_epr cl
    else ()
    );
    (if props.horn then
      props.horn <- Clause.is_horn cl
    else ()
    );
    (if not props.has_eq then
      props.has_eq <- Clause.has_eq_lit cl
    else ());
    (*debug*)
    (* (if not (Clause.is_horn cl) then
       (out_str "\nNon_horn\n";
       flush stdout;
       out_str (Clause.to_tptp cl))
       else ()
       );*)
    (*debug*)
  in
  List.iter f clause_list;
  props

let input_problem_props = ref (empty_problem_props ())

(*-------------Symbol type check--------------------------*)
(* check if there is only on symbol name for each type  *)

module NameSymMap = Map.Make(String)

(* table from symb numbes to list of symbols with the same name *)

type symb_name_table = (((Symbol.symbol list) ref) NameSymMap.t)

let create_name_symb_table () =
  let f symb symb_table =
    if not (Symbol.is_input symb)
    then symb_table
    else
      (
       let symb_name = (Symbol.get_name symb) in
       try
	 let symb_list_ref = NameSymMap.find symb_name symb_table in
	 symb_list_ref:= symb:: (!symb_list_ref);
	 symb_table
       with
	 Not_found ->
	   NameSymMap.add symb_name (ref [symb]) symb_table
      )
  in
  SymbolDB.fold f !symbol_db_ref NameSymMap.empty

let check_symb_name_table snt =
  let ok = ref true in
  let f sname symb_list_ref =
    if ((List.length !symb_list_ref) > 1)
    then
      (
       ok:= false;
       out_str (pref_str^("Type Check Faild on ")^(sname)^"\n");
       List.iter (fun symb ->
	 Symbol.to_stream_full stdout_stream symb;
	 out_str "\n")
	 !symb_list_ref
      )
    else ()
  in
  NameSymMap.iter f snt;
  (if !ok
  then ()
  else
    failwith "Type Check Faild, see help on option --symbol_type_check"
  )

let symb_type_check () =
  let snt = create_name_symb_table () in
  check_symb_name_table snt

let has_equality clause_list = 
  (Clause.has_eq_lit_clause_list clause_list) || (val_of_override !current_options.bmc1_incremental)



(*------Prolific Symbols change for large theories----------------*)

(* If prolific_symb_bound is changed in current_options *)
(* then we need to recalculate which terms/clauses contain prolific symbols *)
let rec change_prolific_symb_input input_clauses =
  let rec change_prolific_symb_term t =
    match t with
    | Term.Fun (symb, args, info) ->
	Term.arg_iter change_prolific_symb_term args;
	Term.assign_has_non_prolific_conj_symb t
    | Term.Var _ -> ()
  in
  let change_prolific_symb_clause c =
    Clause.iter change_prolific_symb_term c;
    Clause.reset_has_non_prolific_conj_symb c
  in
  List.iter change_prolific_symb_clause input_clauses

(*--------------------*)
